#include <windows.h>
#include <sys\timeb.h>

#include "Rewind.h"
#include "Log.h"
#include "SaveStates.h"

Uint8 _rewindInitialized = 0;

static volatile Uint8 _rewind_load_scheduled = 0;

#define _DEBUG_BUFFER_SIZE (256)
static char _debugBuffer[_DEBUG_BUFFER_SIZE];

struct timeb _rewindModuleStartTime;
Uint8 _isPastApplicationInitialization = 0;

// how many chunks back we keep
#define REWIND_CHUNK_COUNT 8
// interval between each chunk
#define REWIND_CHUNK_INTERVAL_MS 500

// time at the start of the application when we do nothing
#define REWIND_APP_INIT_TIME (REWIND_CHUNK_INTERVAL_MS*2)

// index 0 is newest chunk
// last index is oldest chunk
Uint8* _rewindChunkBuffers[REWIND_CHUNK_COUNT];

// time of last chunk capture
struct timeb _rewindLastChunkTimestamp;

void rewind_notify_start_of_frame() {
	if (!_rewindInitialized) {
		return;
	}

	struct timeb now;

	if (!_isPastApplicationInitialization) {
		// we're not past the initialization yet
		
		ftime(&now);

		Uint64 runTimeMs = (Uint64)(1000.0 * (now.time - _rewindModuleStartTime.time)
			+ (now.millitm - _rewindModuleStartTime.millitm));

		if (runTimeMs <= REWIND_APP_INIT_TIME) {
			// do nothing if not enough time has passed
			return;
		}
		else {
			// we are past application initialization
			_isPastApplicationInitialization = 1;
		}
	}

	// we are past application initialization

	ftime(&now);

	Uint64 timeSinceLastChunkMs = (Uint64)(1000.0 * (now.time - _rewindLastChunkTimestamp.time)
		+ (now.millitm - _rewindLastChunkTimestamp.millitm));
	if (timeSinceLastChunkMs < REWIND_CHUNK_INTERVAL_MS) {
		// we do not capture another chunk
		return;
	}

	// we must capture another chunk

	// rotate all chunks to the right
	Uint8* firstChunk = _rewindChunkBuffers[REWIND_CHUNK_COUNT - 1];
	for (int i = REWIND_CHUNK_COUNT - 1; i > 0; i--) {
		_rewindChunkBuffers[i] = _rewindChunkBuffers[i-1];
	}
	_rewindChunkBuffers[0] = firstChunk;

	// create new chunk at index 0, if it's still NULL
	Uint8* buf = _rewindChunkBuffers[0];
	if (!buf) {
		// we need to allocate it
		buf = (Uint8*)malloc(STATE_SIZE * sizeof(Uint8));
		if (!buf) {
			log_write("Warning: could not allocate rewind chunk buffer");
			return;
		}
		_rewindChunkBuffers[0] = buf;
	}

	// dump state into chunk buffer
	savestates_save_to_buffer(_rewindChunkBuffers[0]);
	// save timestamp
	_rewindLastChunkTimestamp = now;
}

void rewind_keyup(SDL_Keysym key) {
	if (!_rewindInitialized) {
		return;
	}

	if (!_isPastApplicationInitialization) {
		return;
	}

	switch (key.sym) {
	case SDLK_BACKSPACE:
		_rewind_load_scheduled = 1;
		break;
	}
}

void _rewind_clear_all_chunks() {
	for (int i = 0; i < REWIND_CHUNK_COUNT; i++) {
		if (_rewindChunkBuffers[i]) {
			free(_rewindChunkBuffers[i]);
			_rewindChunkBuffers[i] = NULL;
		}
	}
}

void rewind_handle_load() {
	if (!_rewind_load_scheduled) {
		return;
	}

	// if the chunk buffer has any empty slots, it means not enough time has
	// passed to get an "oldest" chunk
	if (!_rewindChunkBuffers[REWIND_CHUNK_COUNT - 1]) {
		_rewind_load_scheduled = 0;
		return;
	}

	// perform the load
	savestates_load_from_buffer(_rewindChunkBuffers[REWIND_CHUNK_COUNT - 1]);

	// clear history ...
	_rewind_clear_all_chunks();
	// ... and force a new chunk to be saved
	_rewindLastChunkTimestamp = _rewindModuleStartTime;

	_rewind_load_scheduled = 0;
}

void rewind_start() {
	ftime(&_rewindModuleStartTime);
	_rewindLastChunkTimestamp = _rewindModuleStartTime;

	_rewind_clear_all_chunks();

	_rewindInitialized = 1;
}

void rewind_destroy() {
	if (!_rewindInitialized) {
		return;
	}

	_rewindInitialized = 0;
	_rewind_clear_all_chunks();
}
